"""
API module

.. autosummary::

    Sotong
    cd
    shell
    alert
    logger
"""
import os

from fabric.api import local, lcd, settings



cd = lcd

__all__ = ['Sotong', 'cd', 'shell', 'alert', 'logger']



"""
============
Sotong
============

Sotong is a tiny automation tool like Jenkins, Buildbot, etc...

Based on

- `sched <http://docs.python.org/2/library/sched.html>`_
- `watchdog <https://github.com/gorakhargosh/watchdog>`_
- `fabric <https://github.com/fabric/fabric>`_

>>> from sotong.sotong import Sotong
>>> s = Sotong()
>>> @s.watch('/path/to/project')
... def project_file_modified(event):
...     with s.cd("/path/to/project/test"):
...         out = s.shell("python *.py")
...         if out.failed:
...             s.alert(out.stderr)
...             s.logger.info(out.stderr)
...         else:
...             s.logger.info(out.stdout)
...
>>> @s.interval(10)
... def interval():
...      import datetime
...      print datetime.datetime.now()
...
>>> @s.daily(7, 0, 0)
... def daily_task():
...     print "Good morning"
...
>>> s.run()
"""

__version__ = '0.1.6'
__author__ = 'wak'




import os
import sys
import unittest

sys.path.append(os.path.join(os.path.dirname(__file__), ".."))

import sotong


class TestSotong(unittest.TestCase):

    def test_sotong(self):
        self.assertTrue(sotong.sotong.Sotong)

        s = sotong.sotong.Sotong()

        @s.watch('./')
        def project_file_modified(event):
            pass

        @s.interval(10)
        def interval():
            pass

        @s.daily(10, 34, 56)
        def daily_task():
            pass

    def test_import(self):
        from sotong import api, log, event

if __name__ == '__main__':
    unittest.main()



"""
Core package

.. autosummary::

    Sotong
"""
import os
import datetime
from functools import wraps
import time
import tempfile
import threading
import logging
import webbrowser

from fabric.api import local, lcd, settings

from . import log, event

class Sotong(object):

    def __init__(self, tmp_dir=None, watch_interval=10, autoreload=False):
        if tmp_dir:
            self.tmp_dir = tmp_dir
        else:
            self.tmp_dir = tempfile.mkdtemp()
        self._watch_interval = watch_interval
        self._autoreload = autoreload
        self.logger = log.get_logger()
        self.scheduler = event.Scheduler()
        self.watch_observer = event.WatchObserver()

    def run(self):
        self.logger.addHandler(logging.StreamHandler())
        self.logger.addHandler(log.JSONLogHandler(os.path.join(self.tmp_dir,
                                                               "litebot.log")))
        print "##### START litebot tmp_dir=%s #####" % self.tmp_dir

        self.scheduler_thread = threading.Thread(target=self.scheduler.run, args=[])
        self.scheduler_thread.daemon = True
        self.scheduler_thread.start()
        self.watch_observer.start()
        try:
            while True:
                time.sleep(self._watch_interval)
        except KeyboardInterrupt:
            self.watch_observer.stop()
        self.watch_observer.join()
        print "##### STOP litebot tmp_dir=%s #####" % self.tmp_dir

    def watch(self, path, event_types=None):
        if not event_types:
            event_types = ("modified", )
        def decorator(f):
            @wraps(f)
            def decorated_function(event):
                if event.event_type in event_types:
                    f(event)
            event_handler = event.WatchEventHandler(decorated_function)
            self.watch_observer.schedule(event_handler, path=path, recursive=True)
            return decorated_function
        return decorator

    def interval(self, time):
        def decorator(f):
            @wraps(f)
            def decorated_function():
                self.scheduler.enter(time, 1, decorated_function, ())
                f()
            self.scheduler.enter(time, 1, decorated_function, ())
            return decorated_function
        return decorator

    def daily(self, hour, minutes, seconds):
        def decorator(f):
            @wraps(f)
            def decorated_function():
                f()
            datetime_time = datetime.time(hour, minutes, seconds)
            datetime_datetime = datetime.datetime.combine(datetime.datetime.now(),
                                                          datetime_time)
            daily_time = time.mktime(datetime_datetime.timetuple())
            if daily_time < time.time():
                return
            self.scheduler.enterabs(daily_time, 2, decorated_function, ())
            return decorated_function
        return decorator

    def alert(self, message):
        path = os.path.join(self.tmp_dir, "alert.html")
        with open(path, "w") as f:
            f.write("<pre>%s</pre>" % message)
        webbrowser.open("file://" + path, autoraise=True)

    def shell(self, command):
        with settings(warn_only=True):
            return local(command, capture=True)

    def cd(self, path):
        return lcd(path)



#!/usr/bin/env python

import sys

from setuptools import setup, find_packages
import sotong


long_description = """
Sotong %s

%s
""" % (sotong.__version__, sotong.__doc__.strip())

setup(
    name='sotong',
    version=sotong.__version__,
    description='Tiny automation tool like Jenkins, Buildbot, etc...',
    long_description=long_description,
    author=sotong.__author__,
    packages=find_packages(),
    install_requires=['Fabric>=1.8.0',
                      'watchdog>=0.6.0'],
)



"""
Logging module

.. autosummary::

    JSONLogHandler
    get_logger
"""

import logging
import logging.handlers
import json


class JSONLogHandler(logging.handlers.RotatingFileHandler):
    def format(self, record):
        keys = ("created", "levelname", "pathname", "lineno", "msg", "args",
                "funcName", "exc_info")
        d = {}
        for key in keys:
            d[key] = _jsonize(getattr(record, key))
        return json.dumps(d)


def _jsonize(value):
    """ Jsonize value

    >>> _jsonize(123) + 1
    124
    >>> _jsonize('ABC')
    'ABC'
    >>> _jsonize([123, 'ABC', [1, 2, 3]])
    [123, 'ABC', [1, 2, 3]]
    >>> _jsonize({'a': 'AAA', 'b': 'BBB'})
    {'a': 'AAA', 'b': 'BBB'}
    >>> type(_jsonize(lambda x: x * 2))
    <type 'str'>
    """
    if type(value) in (str, int):
        return value 
    if type(value) in (list, tuple):
        try:
            json.dumps({'value': list(value)})
            return value
        except Exception as e:
            return str(value)
    try:
        json.dumps(value)
        return value
    except Exception as e:
        return str(value)


def get_logger():
    return logging.getLogger("sotong")




"""

.. autosummary::

    Scheduler
    WatchObserver
    WatchEventHandler
"""
import time
from functools import wraps

import sched
import watchdog
import watchdog.observers

from . import log


class Scheduler(object):

    def __init__(self):
        self._scheduler = sched.scheduler(time.time, time.sleep)
        self.logger = log.get_logger()
        self.before_callbacks = []
        self.after_callbacks = []

    def run(self):
        self._scheduler.run()

    def _wrap(self, callback):
        @wraps(callback)
        def wrapped_callback():
            try:
                for before in self.before_callbacks:
                    before()
                callback()
                for after in self.after_callbacks:
                    after()
            except Exception as e:
                self.logger.error(e, exc_info=True)
        return wrapped_callback

    def enter(self, time, priority, callback, args):
        self._scheduler.enter(time, priority, self._wrap(callback), args)

    def enterabs(self, time, priority, callback, args):
        self._scheduler.enterabs(time, priority, self._wrap(callback), args)


class WatchObserver(watchdog.observers.Observer):

    def __init__(self, *args, **kwargs):
        super(WatchObserver, self).__init__(*args, **kwargs)
        self.logger = log.get_logger()
        self.before_callbacks = []
        self.after_callbacks = []

    def dispatch_events(self, event_queue, timeout):
        """ override :py:func:`watchdog.observers.Observer.dispatch_event` """
        event, watch = event_queue.get(block=True, timeout=timeout)
        try:
            for before in self.before_callbacks:
                before(event)
            self._dispatch_event(event, watch)
            for after in self.after_callbacks:
                after(event)
        except Exception as e:
            self.logger.error(e, exc_info=True)
        event_queue.task_done()


class WatchEventHandler(watchdog.events.FileSystemEventHandler):

    def __init__(self, callback, *args, **kwargs):
        super(WatchEventHandler, self).__init__(*args, **kwargs)
        self.callback = callback

    def on_any_event(self, event):
        self.callback(event)

